/* LibCMT - Composable Memory Transactions Library.
 * 
 * Copyright (C) 2005 Duilio Protti <dprotti@users.sourceforge.net>
 *
 * Author: Duilio Protti <dprotti@users.sourceforge.net>
 *
 * $Id:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __G_TRANSACTION_H__
#define __G_TRANSACTION_H__

#include <glib.h>
#include <setjmp.h>

G_BEGIN_DECLS

typedef struct _GTLog GTLog;
struct _GTLog {
  /*< private >*/
  GSList *tvar_entries;
};

typedef struct _GTVarWaitQueue GTVarWaitQueue;
struct _GTVarWaitQueue {
  /*< private >*/
  GCond *wait_queue;
  GTLog *parent_tlog;
};

/**
 * @addtogroup GTVar GTVar
 * @{
 */

/**
 * Type of a function that duplicates a shared variable (not
 * a transactional variable).
 *
 * @param svar Pointer to the shared variable.
 *
 * @return A copy of the shared variable into a newly allocated
 * memory.
 */
typedef gpointer (*GTVarDupFunc)  (gconstpointer svar);

/**
 * Type of a function that copy a contiguous shared variable (not
 * a transactional variable) from @a src to @a dst.
 */
typedef void     (*GTVarCopyFunc) (gpointer dst, gconstpointer src);

/**
 * Type representing a transactional variable which encapsulates a
 * shared variable.
 */
typedef struct _GTVar GTVar;

/**
 * @}
 */

struct _GTVar {
  /*< private >*/
  gpointer var;

  GTVarDupFunc dup;
  GDestroyNotify destroy;
  GTVarCopyFunc copy;
  GCompareFunc compare;

  GSList *wait_queue_list;

  gsize size;
  unsigned int is_contiguous : 1;
};

typedef struct _GTVarEntry GTVarEntry;
struct _GTVarEntry {
  GTVar *tvar;
  gpointer old_value;
  gpointer new_value;

  unsigned int was_readed : 1;
};

/**
 * @addtogroup GTransaction GTransaction
 * @{
 */

/**
 * Type representing a transaction.
 */
typedef struct _GTransaction GTransaction;

/**
 * Type of a function which performs the actions of a transaction.
 *
 * When invoked, @a tr will be the transaction on which the function is
 * involved.
 */
typedef void (*GTransactionFunc) (GTransaction *tr, gpointer user_data);

/**
 * @}
 */

struct _GTransaction {
  /*< private >*/
  gchar            *name;
  GTLog            *log;
  GTransactionFunc func;
  gpointer         user_data;
  sigjmp_buf       jmp_buf;
  GTransaction     *nested;
  GTransaction     *enclosing;
  GTransaction     *next;
};

GTVar*        g_tvar_new                  (gpointer var,
                                           GTVarDupFunc dup_func,
                                           GDestroyNotify destroyer,
                                           GCompareFunc compare);
GTVar*        g_tvar_new_contiguous       (gpointer var,
                                           gsize size,
                                           GTVarCopyFunc copy_func,
                                           GCompareFunc compare);
gpointer      g_transaction_read_tvar     (GTransaction *transaction,
                                           GTVar *tvar);
void          g_transaction_add_tvar      (GTransaction *transaction,
                                           GTVar *var);

GTransaction* g_transaction_new           (const gchar *name);
void          g_transaction_set_func      (GTransaction *transaction,
                                           GTransactionFunc func,
                                           gpointer user_data);
void          g_transaction_do            (GTransaction *transaction);
gboolean      g_transaction_timed_do      (GTransaction *transaction,
                                           const GTimeVal *abs_time);
void          g_transaction_retry         (GTransaction *transaction);
GTransaction* g_transaction_sequence      (GTransaction *tr1, GTransaction *tr2);
GTransaction* g_transaction_or_else       (GTransaction *tr1, GTransaction *tr2);
GTransaction* g_transaction_copy          (GTransaction *transaction);
void          g_transaction_destroy       (GTransaction *transaction);

G_END_DECLS

#endif /* __G_TRANSACTION_H__ */
